#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <utime.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/vfs.h>

#include <linux/types.h>
#include <linux/dirent.h>
#include <linux/unistd.h>

#include "CFSimpl.h"

_syscall3(int, getdents, uint, fd, struct dirent *, dirp, uint, count);

namespace CFS {

    class VirtualSB : public virtual POA_CFS::Superblock {

	public:
	VirtualSB(CORBA::Object_var content, CORBA::ORB_var orb) {
	    inode = new Inode;

	    inode->i_ino = random();
	    inode->i_nlink = 1;
	    inode->i_dev = 0;
	    inode->i_uid = inode->i_gid = inode->i_rdev = 0;
	    inode->i_size = inode->i_blksize = inode->i_blocks = 0;
	    inode->i_atime = inode->i_mtime = inode->i_ctime = time(NULL);
	    inode->i_sb = CFS::Superblock::_duplicate(this->_this());
	
	    inode->i_mode = S_IROTH | S_IRGRP | S_IRUSR | S_IFREG;
	    inode->i_size = 269;
	    this->content = content;
	    this->orb = orb;
	}

	VirtualSB(CORBA::UShort mode, CORBA::UShort dev, const char * name) {
	    inode = new Inode;

	    inode->i_ino = random();
	    inode->i_nlink = 1;
	    inode->i_dev = 0;
	    inode->i_uid = inode->i_gid = inode->i_rdev = 0;
	    inode->i_size = inode->i_blksize = inode->i_blocks = 0;
	    inode->i_atime = inode->i_mtime = inode->i_ctime = time(NULL);
	    inode->i_sb = CFS::Superblock::_duplicate(this->_this());
	
	    inode->i_name = CORBA::string_dup(name);
	    inode->i_dev = dev;
	    inode->i_mode = mode;
	    inode->i_size = 0;
	    this->memory = new vector<char>;
	}

	CORBA::Long read_inode(const char *path, Inode_out inode) {
	    inode = new Inode(*this->inode);
	    return 0;
	}

	CORBA::Long read_page(const char* path, CORBA::ULong offset, Page_out page) {
	    if(!CORBA::is_nil(content)) {
		char *IOR = orb->object_to_string(content);

		page.nbytes = 0;

		if (offset > strlen(IOR))
		    return 0;

		memcpy(page.data, IOR + offset, strlen(IOR) - offset);
		page.nbytes = strlen(IOR) - offset + 1;
		page.data[page.nbytes-1] = '\n';

		return 0;       // devolver a IOR
	    }

	    // se mt_object  null, temos um arquivo virtual (memria do CFSd)
	    page.nbytes = 0;
	    while ((page.nbytes < PAGE_SIZE) && (offset < memory->size())) {
		page.data[page.nbytes] = (*memory)[offset];
		page.nbytes++;
		offset++;
	    }

	    return 0;
	}

	CORBA::Long commit_write(const char* path, CORBA::ULong offset, const Page& page) {
	    if(!CORBA::is_nil(content))
		return EINVAL;

	    //	    if (offset + page.nbytes > memory->size())
	    memory->resize(offset + page.nbytes);
	    
	    for(unsigned i = 0; i < page.nbytes; i++)
		(*memory)[offset + i] = page.data[i];

	    this->inode->i_size = memory->size();
	    //this->my_inode->i_atime = this->my_inode->i_mtime = this->my_inode->i_ctime = time(NULL);

	    return 0;
	}

	CORBA::Long read_link(const char* path, Page_out contents) { 
	    if(!CORBA::is_nil(content))
		return EINVAL; 

	    if (!S_ISLNK(inode->i_mode))
		return EINVAL;

	    //contents = new Page;
	    contents.nbytes = memory->size();
	    for(unsigned i = 0; i < memory->size(); i++)
		contents.data[i] = (*memory)[i];
	    
	    return 0;
	}

	CORBA::Long chown(const char* path, CORBA::Long uid, CORBA::Long gid) {
	    if(!CORBA::is_nil(content))
		return EINVAL; 
	    
	    if (uid > -1) 
		inode->i_uid = uid;
	    
	    if (gid > -1)
		inode->i_gid = gid;

	    return 0;
	}

	CORBA::Long chmod(const char* path, CORBA::UShort mode) {
	    if(!CORBA::is_nil(content))
		return EINVAL; 
	    
	    inode->i_mode = mode;

	    return 0;
	}

	CORBA::Long utime(const char* path, CORBA::Long atime, CORBA::Long mtime) {
	    if(!CORBA::is_nil(content))
		return EINVAL; 
	    
	    inode->i_atime = atime;
	    inode->i_mtime = mtime;

	    return 0;
	}

	CORBA::Long statfs(const char *path, Statfs& buf)                              { return ENOSYS;  }
	CORBA::Long symlink(const char* oldpath, const char* newpath)                  { return ENOTDIR; }
	CORBA::Long readdir(const char* path, Dirents_out dir)                         { return ENOTDIR; }
	CORBA::Long mkdir(const char* path, CORBA::UShort mode)                        { return ENOTDIR; }
	CORBA::Long mknod(const char* path, CORBA::UShort mode, CORBA::UShort dev)     { return ENOTDIR; }
	CORBA::Long unlink(const char* path)                                           { return ENOTDIR; }
	CORBA::Long rmdir(const char* path)                                            { return ENOTDIR; }
	CORBA::Long rename(const char* oldpath, const char* newpath)                   { return ENOTDIR; }

	private:
	CORBA::ORB_var orb;
	CORBA::Object_var content;
	// case of virtual file
	vector<char> *memory;
	Inode *inode;
    };

    class ContextSB : public virtual POA_CFS::Superblock
    {
	public:
	ContextSB(CosNaming::NamingContext_var root, CORBA::ORB_var orb) {
	    this->orb = CORBA::ORB::_duplicate(orb);
	    this->root = root;

	    inode = new Inode;
	    inode->i_name = CORBA::string_dup("");
	    inode->i_ino = random();
	    inode->i_nlink = 1;
	    inode->i_dev = inode->i_uid = inode->i_gid = inode->i_rdev = 0;
	    inode->i_size = inode->i_blksize = inode->i_blocks = 0;
	    inode->i_atime = inode->i_mtime = inode->i_ctime = time(NULL);
	    inode->i_mode = S_IROTH | S_IRGRP | S_IRUSR | S_IXOTH | S_IXGRP | S_IXUSR | S_IWUSR | S_IFDIR;
	    inode->i_sb = CFS::Superblock::_duplicate(this->_this());
	}

	// file operations

	CORBA::Long readdir(const char *path, Dirents_out dir) {
	    // if path is not null, look for the right Inode
	    if (strcmp(path, "") != 0) {
		Inode_var inode;
		
		int error = read_inode(path, inode);

		if (error)
		    return error;
		
		return inode->i_sb->readdir(inode->i_name, dir);
	    }

	    CosNaming::BindingIterator_var it;
	    CosNaming::BindingList_var blist;

	    root->list(100, blist, it);

	    dir = new Dirents;
	    do {
		dir->length(blist->length());
		for (CORBA::ULong i = 0; i < blist->length(); i++) {
		    Dirent direntry;
		    direntry.d_name = CORBA::string_dup(blist[i].binding_name[0].id);
		    direntry.d_ino  = i + 6660000;
		    dir[i] = direntry;
		}
	    } while(!CORBA::is_nil(it) && it->next_n(100, blist));

	    if (!CORBA::is_nil(it))
		it->destroy();
	    
	    return 0;
	}

	CORBA::Long statfs(const char *path, Statfs& buf) {
	    if (strcmp(path, "") != 0) {
		Inode_var inode;
		
		int error = read_inode(path, inode);
		
		if (error)
		    return error;
		
		return inode->i_sb->statfs(inode->i_name, buf);
	    }
	    
	    return ENOSYS;
	}

	CORBA::Long read_inode(const char *path, Inode_out inode) {

	    if (CORBA::is_nil(root) || !strcmp(path, "")) {
		inode = new Inode(*this->inode);
		return 0;
	    }

	    CosNaming::Name cosname;
	    path_tokenizer(path, cosname);
	    
	    return read_inode(cosname, inode);
	}

	CORBA::Long read_page(const char* path, CORBA::ULong offset, Page& page) {
	    if (strcmp(path, "") != 0) {
		Inode_var inode;
		
		int error = read_inode(path, inode);
		
		if (error)
		    return error;
		
		return inode->i_sb->read_page(inode->i_name, offset, page);
	    }
	    
	    return EISDIR;
	}
	
	CORBA::Long read_link(const char* path, Page& page) {
	    if (strcmp(path, "") != 0) {
		Inode_var inode;
		
		int error = read_inode(path, inode);
		
		if (error)
		    return error;
		
		return inode->i_sb->read_link(inode->i_name, page);
	    }
	    
	    return EISDIR;
	}

	CORBA::Long commit_write(const char* path, CORBA::ULong offset, const Page& page) {
	    if (strcmp(path, "") != 0) {
		Inode_var inode;

		int error = read_inode(path, inode);

		if (error)
		    return error;
		
		return inode->i_sb->commit_write(inode->i_name, offset, page);
	    }

	    return EISDIR;
	}

	CORBA::Long mkdir(const char* path, CORBA::UShort mode) {
	    CosNaming::Name cosname;

	    path_tokenizer(path, cosname);

	    // if path is not 1 word, look for the right Inode
	    if (cosname.length() > 1) {
		Inode_var inode;
		char name[1024];

		//strcpy(name, cosname[cosname.length() - 1].id);
		CosNaming::NameComponent dir = cosname[cosname.length() - 1];
		cosname.length(cosname.length() - 1);
		
		int error = read_inode(cosname, inode);
		
		if (error)
		    return error;
		
		name[0] = 0;
		strcat(name, inode->i_name);
		strcat(name, "/");
		strcat(name, dir.id);
		return inode->i_sb->mkdir(name, mode);
	    }

	    try {
		root->bind_new_context(cosname);
	    }
	    catch(CosNaming::NamingContext::AlreadyBound e) {
		return EEXIST;
	    }

	    return 0;
	}

	CORBA::Long mknod(const char* path, CORBA::UShort mode, CORBA::UShort dev) {
	    CosNaming::Name cosname;
	    
	    path_tokenizer(path, cosname);

	    // if path is not 1 word, look for the right Inode
	    if (cosname.length() > 1) {
		Inode_var inode;
		char name[1024];

		//strcpy(name, cosname[cosname.length() - 1].id);
		CosNaming::NameComponent nod = cosname[cosname.length() - 1];
		cosname.length(cosname.length() - 1);
		
		int error = read_inode(cosname, inode);

		if (error)
		    return error;

		name[0] = 0;
		strcat(name, inode->i_name);
		strcat(name, "/");
		strcat(name, nod.id);

		return inode->i_sb->mknod(name, mode, dev);
	    }

	    try {
		root->bind(cosname, createVirtualSB(mode, dev, path));
	    }
	    catch(CosNaming::NamingContext::AlreadyBound e) {
		return EEXIST;
	    }
	    
	    return 0;
	}

	CORBA::Long unlink(const char* path) { 
	    CosNaming::Name cosname;
	    
	    path_tokenizer(path, cosname);

	    // if path is not 1 word, look for the right Inode
	    if (cosname.length() > 1) {
		Inode_var inode;
		char name[1024];

		//strcpy(name, cosname[cosname.length() - 1].id);
		CosNaming::NameComponent nod = cosname[cosname.length() - 1];
		cosname.length(cosname.length() - 1);
		
		int error = read_inode(cosname, inode);

		if (error)
		    return error;
		
		name[0] = 0;
		strcat(name, inode->i_name);
		strcat(name, "/");
		strcat(name, nod.id);

		return inode->i_sb->unlink(name);
	    }

	    try {
		CORBA::Object_var obj = root->resolve(cosname);

		CosNaming::NamingContext_var dir = CosNaming::NamingContext::_narrow(obj);
		if (!CORBA::is_nil(dir))
		    return EISDIR; //dir->destroy();

		root->unbind(cosname);
	    }
	    catch(CosNaming::NamingContext::NotFound e) {
		return ENOENT;
	    }
	    catch(CORBA::Exception e) {
		return EPERM;
	    }

	    return 0;
	}

	CORBA::Long rmdir(const char* path) { 
	    CosNaming::Name cosname;
	    
	    path_tokenizer(path, cosname);

	    // if path is not 1 word, look for the right Inode
	    if (cosname.length() > 1) {
		Inode_var inode;
		char name[1024];

		//strcpy(name, cosname[cosname.length() - 1].id);
		CosNaming::NameComponent nod = cosname[cosname.length() - 1];
		cosname.length(cosname.length() - 1);
		
		int error = read_inode(cosname, inode);

		if (error)
		    return error;
		
		name[0] = 0;
		strcat(name, inode->i_name);
		strcat(name, "/");
		strcat(name, nod.id);

		return inode->i_sb->rmdir(name);
	    }

	    try {
		CORBA::Object_var obj = root->resolve(cosname);
	    
		CosNaming::NamingContext_var dir = CosNaming::NamingContext::_narrow(obj);
		if (CORBA::is_nil(dir))
		    return ENOTDIR;

		dir->destroy();
		root->unbind(cosname);
	    }
	    catch(CosNaming::NamingContext::NotFound e) {
		return ENOENT;
	    }
	    catch(CosNaming::NamingContext::NotEmpty e) {
		return ENOTEMPTY;
	    }
	    catch(CORBA::Exception e) {
		return EPERM;
	    }

	    return 0; 
	}

	CORBA::Long rename(const char* oldpath, const char* newpath) { 
	    CosNaming::Name cosnameold;
	    CosNaming::Name cosnamenew;

	    path_tokenizer(oldpath, cosnameold);
	    path_tokenizer(newpath, cosnamenew);
	    
	    // if path is not 1 word, look for the right Inode
	    if (cosnameold.length() > 1) {
		Inode_var inode;
		char nameold[1024], namenew[1024];

		//strcpy(name, cosname[cosname.length() - 1].id);
		CosNaming::NameComponent nodold = cosnameold[cosnameold.length() - 1];
		CosNaming::NameComponent nodnew = cosnamenew[cosnameold.length() - 1];

		cosnameold.length(cosnameold.length() - 1);
		
		int error = read_inode(cosnameold, inode);
		
		if (error)
		    return error;
		
		nameold[0] = 0;
		strcat(nameold, inode->i_name);
		strcat(nameold, "/");
		strcat(nameold, nodold.id);

		namenew[0] = 0;
		strcat(namenew, inode->i_name);
		strcat(namenew, "/");
		strcat(namenew, nodnew.id);

		return inode->i_sb->rename(nameold, namenew);
	    }

	    try {
		CORBA::Object_var obj = root->resolve(cosnameold);
	    
		CosNaming::NamingContext_var dir = CosNaming::NamingContext::_narrow(obj);
		
		if (CORBA::is_nil(dir))
		    root->bind_context(cosnamenew, dir);
		else
		    root->bind(cosnamenew, obj);

		root->unbind(cosnameold);
	    }
	    catch(CosNaming::NamingContext::NotFound e) {
		return ENOENT;
	    }
	    catch(CORBA::Exception e) {
		return EPERM;
	    }
	    
	    return 0; 
	}

	CORBA::Long symlink(const char* oldpath, const char* newpath) {
	    CosNaming::Name cosname;
	    
	    path_tokenizer(newpath, cosname);

	    // if path is not 1 word, look for the right Inode
	    if (cosname.length() > 1) {
		Inode_var inode;
		char name[1024];

		//strcpy(name, cosname[cosname.length() - 1].id);
		CosNaming::NameComponent nod = cosname[cosname.length() - 1];
		cosname.length(cosname.length() - 1);
		
		int error = read_inode(cosname, inode);
		
		if (error)
		    return error;
		
		name[0] = 0;
		strcat(name, inode->i_name);
		strcat(name, "/");
		strcat(name, nod.id);

		return inode->i_sb->symlink(oldpath, name);
	    }

	    int error = 0;
	    error = mknod(newpath, S_IFLNK | S_IROTH | S_IRGRP | S_IRUSR | S_IXOTH | S_IXGRP | S_IXUSR | S_IWUSR, 0);

	    if (error)
		return error;
	    
	    Inode_var inode;
	    error = read_inode(newpath, inode);

	    if (error)
		return error;
	    
	    Page page;
	    page.nbytes = strlen(oldpath);

	    memcpy(page.data, oldpath, page.nbytes+1);

	    error = inode->i_sb->commit_write(newpath, 0, page);

	    if (error)
		return error;

	    return 0;
	}
	
	CORBA::Long chown(const char* path, CORBA::Long uid, CORBA::Long gid) {
	    if (strcmp(path, "") != 0) {
		Inode_var inode;
		
		int error = read_inode(path, inode);
		
		if (error)
		    return error;
		
		return inode->i_sb->chown(inode->i_name, uid, gid);
	    }

	    return EPERM; 
	}
	
	CORBA::Long chmod(const char* path, CORBA::UShort mode) { 
	    if (strcmp(path, "") != 0) {
		Inode_var inode;
		
		int error = read_inode(path, inode);
		
		if (error)
		    return error;
		
		return inode->i_sb->chmod(inode->i_name, mode);
	    }

	    return EPERM;
	}
	
	CORBA::Long utime(const char* path, CORBA::Long atime, CORBA::Long mtime) { 
	    if (strcmp(path, "") != 0) {
		Inode_var inode;
		
		int error = read_inode(path, inode);

		if (error)
		    return error;
		
		return inode->i_sb->utime(inode->i_name, atime, mtime);
	    }

	    return EPERM;
	}

	private:
	CORBA::ORB_var orb;
	CosNaming::NamingContext_var root;
	Inode *inode;

	CORBA::Long read_inode(CosNaming::Name &cosname, Inode_out inode) {
	    CORBA::Object_var obj;
	    Superblock_var sb;
	    char path[1024];
	    path[0] = 0;

	    if (cosname.length() == 0) 
		return read_inode("", inode);

	    CosNaming::Name name;
	    try {
		name.length(1);
		name[0] = cosname[0];

		obj = root->resolve(name);

		for (unsigned i = 0; i < cosname.length() - 1; i++)
		    cosname[i] = cosname[i + 1];
		cosname.length(cosname.length() - 1);
		
		CosNaming::NamingContext_var context = CosNaming::NamingContext::_narrow(obj);

		for (unsigned i = 0; i < cosname.length(); i++) {
		    strcat(path, "/");
		    strcat(path, cosname[i].id);
		}
		
		if(!CORBA::is_nil(context)) {
		    sb = createContextSB(context, this->orb);
		    return sb->read_inode(path, inode);
		}

		sb = CFS::Superblock::_narrow(obj);

		if(CORBA::is_nil(sb)) {
		    sb = createVirtualSB(obj, this->orb);
		    return sb->read_inode("", inode);
		}

	    }
	    catch(CosNaming::NamingContext::NotFound e) {
		switch(e.why) {
		case CosNaming::NamingContext::missing_node:
		    return ENOENT;
		case CosNaming::NamingContext::not_context:
		    sb = CFS::Superblock::_narrow(obj);
		    
		    if(CORBA::is_nil(sb))
			return ENOTDIR;
		    
		    for (unsigned i = 0; i < e.rest_of_name.length(); i++) {
			strcat(path, "/");
			strcat(path, e.rest_of_name[i].id);		
		    }

		    break;
		case CosNaming::NamingContext::not_object:
		    return ENOENT;
		}		
	    }
	    catch(CosNaming::NamingContext::CannotProceed e) {
		
		
		return EFAULT;
	    }
	    catch(CosNaming::NamingContext::InvalidName e) {
		return EACCES;
	    }
	    
	    return sb->read_inode(CORBA::string_dup(path), inode);
	}
    };

    class DiskSB : public virtual POA_CFS::Superblock 
    {
	public:	

	DiskSB(const char *dev, long mode) {
	    this->dev = CORBA::string_dup(dev);
	    this->mode = mode;
	}

	CORBA::Long statfs(const char *path, Statfs& buf) {
	    struct statfs mybuf;
	    char local[2048];

	    local[0] = 0;
	    strcat(local, dev);
	    strcat(local, path);

	    if(::statfs(local, &mybuf) == -1) {
		cout << "(ServerCFS) Superblock statfs error: " << strerror(errno) << endl;
		return errno;
	    }

	    buf.f_type    = mybuf.f_type;
	    buf.f_bsize   = mybuf.f_bsize;
	    buf.f_blocks  = mybuf.f_blocks;
	    buf.f_bfree   = mybuf.f_bfree;
	    buf.f_bavail  = mybuf.f_bavail;
	    buf.f_files   = mybuf.f_files;
	    buf.f_ffree   = mybuf.f_ffree;
	    memcpy (&buf.f_fsid, &mybuf.f_fsid, 8);
	    buf.f_namelen = mybuf.f_namelen;

	    return 0;
	}
	
	CORBA::Long read_inode(const char *path, Inode_out inode) {
	    struct stat buf;
	    char local[2048];

	    local[0] = 0;
	    strcat(local, dev);
	    strcat(local, CORBA::string_dup(path));

	    if(lstat(local, &buf) == -1) {
		inode = NULL;
		cout << "(ServerCFS) Superblock read_inode error: " << strerror(errno) << endl;
		cout << "(ServerCFS) Was trying to find: " << local << endl;
		return errno;
	    }

	    Inode *my_inode = new Inode;
	    my_inode->i_name    = CORBA::string_dup(path);
	    my_inode->i_dev     = buf.st_dev;
	    my_inode->i_ino     = buf.st_ino;
	    my_inode->i_mode    = buf.st_mode;
	    my_inode->i_nlink   = buf.st_nlink;
	    my_inode->i_uid     = buf.st_uid;
	    my_inode->i_gid     = buf.st_gid;
	    my_inode->i_rdev    = buf.st_rdev;
	    my_inode->i_size    = buf.st_size;
	    my_inode->i_blksize = buf.st_blocks;
	    my_inode->i_atime   = buf.st_atime;
	    my_inode->i_mtime   = buf.st_mtime;
	    my_inode->i_ctime   = buf.st_ctime;
	    my_inode->i_sb      = CFS::Superblock::_duplicate(this->_this());

	    inode = my_inode;
	    return 0;
	}

	CORBA::Long readdir(const char *path, Dirents_out dir) {
	    int fd_dir;
	    char local[1024];
	    local[0] = 0;
	    strcat(local, this->dev);
	    strcat(local, path);
	    if ((fd_dir = open(local, O_RDONLY)) == -1)
		return errno;

	    struct dirent *dirp = (struct dirent*) malloc(1024*sizeof(struct dirent));
	    CORBA::Long nbytes = 0;
	    nbytes = ::getdents(fd_dir, dirp, 1024*sizeof(struct dirent));
	    close(fd_dir);
	    
	    if (nbytes == -1) {
		free(dirp);
		return errno;
	    }
	    
	    dir = new Dirents;
	    struct dirent *dirlist = dirp;
	    while (nbytes > 0) {
		Dirent direntry;

		direntry.d_ino = dirlist->d_ino;
		direntry.d_name = CORBA::string_dup(dirlist->d_name);

		dir->length(dir->length()+1);
		dir[dir->length()-1] = direntry;

		nbytes -= dirlist->d_reclen;

		dirlist = (struct dirent *) ((char*) dirlist + dirlist->d_reclen);
	    }

	    return 0;
	}

        /* L uma pgina do arquivo especificado pelo caminho relativo path, a partir de offset, e coloca em page */
        CORBA::Long read_page(const char* path, CORBA::ULong offset, Page_out page) {
            int fd_file;
            char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, path);

            printf("Read_page name=%s, offset=%ld\n", local, offset);

            if ((fd_file = open(local, O_RDONLY)) == -1) {
		cout << "Read_page error: " << strerror(errno) << endl;   
                return errno;
            }

            if (lseek(fd_file, offset, SEEK_SET) == -1) {
                cout << "Read_page error:" << strerror(errno) << endl;
                return errno;
            }
		
	    //page = new Page;
            //page->data = (char*) malloc(sizeof(char)*PAGE_SIZE);
	    int error = 0;
            if ((error = read(fd_file, page.data, PAGE_SIZE)) == -1) {
                cout << "Read_page error: "  << strerror(errno) << endl;
                return errno;
            }
	    page.nbytes = error;		
            close(fd_file);
			
            return 0;
        }

        /* Escreve uma pgina no arquivo especificado pelo caminho relativo path, a partir de offset, usando os dados de page */
        CORBA::Long commit_write(const char* path, CORBA::ULong offset, const Page& page) {
            int fd_file;
            char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, path);

            //            printf("Commit_write name=%s, offset=%ld, nbytes=%ld, DATA=%s\n", local, offset, page->nbytes, page->data);

            if ((fd_file = open(local, O_WRONLY)) == -1) {
		cout << "Commit_write error:" << strerror(errno) << endl;
                return errno;
            }

            if (lseek(fd_file, offset, SEEK_SET) == -1) {
		cout << "Commit_write error:" << strerror(errno) << endl;
                return errno;
            }
			
            if (write(fd_file, page.data, page.nbytes) == -1) {
		cout << "Commit_write error:" << strerror(errno) << endl;
                return errno;
            }
			
            close(fd_file);
			
            return 0;
        }

	CORBA::Long mkdir(const char* path, CORBA::UShort mode) {
	    char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, path);

	    if (::mkdir(local, mode) == -1) {
		cout << "(ServerCFS) Mkdir error: " << strerror(errno) << endl;
		return errno;
	    }
	    return 0;
	}

	CORBA::Long mknod(const char* path, CORBA::UShort mode, CORBA::UShort dev) {
	    char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, path);

	    if (::mknod(local, mode, dev) == -1) {
		cout << "(ServerCFS) Mknod error: " << strerror(errno) << endl;
		return errno;
	    }
	    return 0;
	}

	CORBA::Long unlink(const char* path) {
	    char local[1024];
	    local[0] = 0;
	    strcat(local, this->dev);
            strcat(local, path);

	    if (::unlink(local) == -1) {
		cout << "(ServerCFS) Unlink error: " << strerror(errno) << endl;
		return errno;
	    }
	    return 0;
	}

	CORBA::Long rmdir(const char* path) {
	    char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, path);

	    if (::rmdir(local) == -1) {
		cout << "(ServerCFS) Rmdir error: " << strerror(errno) << endl;
		return errno;
	    }
	    return 0;
	}

	CORBA::Long rename(const char* oldpath, const char* newpath) {
	    char local[1024], oldlocal[1024];
            local[0] = oldlocal[0] = 0;
            strcat(local, this->dev);
            strcat(local, newpath);
            strcat(oldlocal, this->dev);
            strcat(oldlocal, oldpath);
	    
	    cout << "RENAME : " << oldlocal << " -> " << local;
	    if (::rename(oldlocal, local) == -1) {
		cout << "(ServerCFS) Rename error: " << strerror(errno) << endl;
		return errno;
	    }
	    return 0;
	}

	CORBA::Long read_link(const char* path, Page_out contents) {
            char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, path);

	    //contents = new Page;
            //contents->data = (char*) malloc(sizeof(char)*PAGE_SIZE);

	    int bytes = 0;
            if ((bytes = readlink(local, (char *)contents.data, PAGE_SIZE)) == -1) {
                cout << "(ServerCFS) Read_link error: "  << strerror(errno) << endl;
                return errno;
            }
	    contents.nbytes = bytes;
	    return 0;
	}

	CORBA::Long symlink(const char* oldpath, const char* newpath) {
	    char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, newpath);
	    
	    if (::symlink(oldpath, local) == -1) {
		cout << "(ServerCFS) Symlink error: " << strerror(errno) << endl;
		cout << "(ServerCFS) Name: " << local << endl;
		return errno;
	    }
	    return 0;
	}

	CORBA::Long chown(const char* path, CORBA::Long uid, CORBA::Long gid) {
	    char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, path);
	    
	    if (::chown(local, uid, gid) == -1) {
		cout << "(ServerCFS) Chmod error: " << strerror(errno) << endl;
		return errno;
	    }
	    return 0;
	}

	CORBA::Long chmod(const char* path, CORBA::UShort mode) {
	    char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, path);
	    
	    if (::chmod(local, mode) == -1) {
		cout << "(ServerCFS) Chmod error: " << strerror(errno) << endl;
		return errno;
	    }
	    return 0;
	}

	CORBA::Long utime(const char* path, CORBA::Long atime, CORBA::Long mtime) {
	    char local[1024];
            local[0] = 0;
            strcat(local, this->dev);
            strcat(local, path);

	    struct utimbuf timebuf;
	    timebuf.actime  = atime;
	    timebuf.modtime = mtime;
	    if (::utime(local, &timebuf) == -1) {
		cout << "(ServerCFS) Utime error: " << strerror(errno) << endl;
		return errno;
	    }
	    return 0;
	}

	private:
	char *dev;
	long mode;
    };

    Superblock_var createDiskSB(char *path, CORBA::UShort mode) {
	DiskSB *sb = new DiskSB(path, mode);
	return Superblock::_duplicate(sb->_this());
    }

    Superblock_var createContextSB(CosNaming::NamingContext_var root, CORBA::ORB_var orb) {
	ContextSB *sb = new ContextSB(root, orb);
	return Superblock::_duplicate(sb->_this());
    }
    
    Superblock_var createVirtualSB(CORBA::Object_var content, CORBA::ORB_var orb) {
	VirtualSB *sb = new VirtualSB(content, orb);
	return Superblock::_duplicate(sb->_this());
    }

    Superblock_var createVirtualSB (CORBA::UShort mode, CORBA::UShort dev, const char * name) {
    	VirtualSB *sb = new VirtualSB(mode, dev, name);
	return Superblock::_duplicate(sb->_this());
    }

    void path_tokenizer(const char *path, CosNaming::Name &cosname) {
	char * name, * end, * mypath;
	name = end = mypath = CORBA::string_dup(path);

	while (*end != '\0') {
	    if (*end == '/') *end = '\0';
	    end++;
	}
	   
	while (*name == '\0') name++;
	while (name < end ) {
	    cosname.length(cosname.length() + 1);
	    cosname[cosname.length() - 1].id = CORBA::string_dup(name);

	    // procuramos pelo prximo nome
	    while (*name != '\0') name++;
	    while (name < end && *name == '\0') name++;
	}
	   
	CORBA::string_free(mypath);
    }
};
