/* 
   Projeto de MAC 5759
   
   Alunos: Livio Baldini Soares <livio@ime.usp.br>                  No. USP: 2869100
   Mrcio Rodrigo de Freitas Carneiro <carneiro@ime.usp.br> No. USP: 2868836
   Roberto Pires de Carvalho <carvalho@ime.usp.br>          No. USP: 2869559
   Data: 5/7/2002
*/


#include <mico/CosNaming.h>
#include <iostream.h>
#include <stack.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/utsname.h>

#include "CFS.h"
#include "CFSimpl.h"

#define EXPORTS "./etc/cfs.exports"

namespace CFS {

    class ServerCFS_impl : public virtual POA_CFS::ServerCFS 
    {
	public:
	  
	ServerCFS_impl() {
	    this->orb = NULL;
	}

	ServerCFS_impl(CORBA::ORB_var orb) {
	    FILE * config = fopen(EXPORTS, "r");
	    char path[2048];
	    char export_path[2048];
	    char options[2048];

	    try {    
		// Initialize orb
		this->orb = orb;

		// Get reference to Root POA.
		CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
		PortableServer::POA_var poa = PortableServer::POA::_narrow(obj);
		
		// Activate POA manager
		poa->the_POAManager()->activate();

		obj = orb->resolve_initial_references("NameService");
		CosNaming::NamingContext_var context = CosNaming::NamingContext::_narrow(obj);
		assert(!CORBA::is_nil(context));   
		superblock = createContextSB(context, orb);

		struct utsname host;
		uname(&host);
 		char local[1024];
		local[0] = 0;
		strcat(local, host.sysname);
		strcat(local, ":");
		strcat(local, host.nodename);
		strcat(local, ":");
                strcat(local, host.release);
		strcat(local, ":");
		strcat(local, host.machine);
									     
		CosNaming::Name kama;
		kama.length(1);
		kama[0].id = local;
		context->bind(kama, CFS::ServerCFS::_duplicate(this->_this())); 
		
		cout << "(ServerCFS) Bindinf myself as " << local << "..." << endl;
		while (!feof(config)) {
		    fscanf(config, "%s", path);
		    fscanf(config, "%s", export_path);
		    fscanf(config, "%s", options);
		    if (path[0] && export_path[0] && options[0]) {
			// criamos o bloco
			struct stat buf;
			Superblock_var sb;
				    
			if (!stat(path, &buf)) {
			    cout << "(ServerCFS) Exporting " << path << " to [/]";

			    CosNaming::Name cosname;
			    path_tokenizer(export_path, cosname);
			    
			    sb = superblock;
			    Inode_var inode;
			    for (unsigned i = 0; i < cosname.length() - 1; i++) {
				sb->mkdir(cosname[i].id, 0);
				sb->read_inode(cosname[i].id, inode);
				sb = inode->i_sb;
				cout << " -> " << cosname[i].id;
			    }

			    sb = createDiskSB(path, buf.st_mode);
			    context->rebind(cosname, Superblock::_duplicate(sb));

			    cout << " -> " << cosname[cosname.length() - 1].id << endl;

			    exports.push_back(CORBA::string_dup(export_path));
			}
			else {
			    printf("Invalid export path: %s\n", path);
			    printf("Error: %s\n", strerror(errno));
			}
		    }
		    path[0] = export_path[0] = options[0] = 0;
		}
		cout << endl;
		fclose(config);
	    }
	    catch (CORBA::Exception e) {
		e._print(cerr);
		cerr << endl;
		cerr << "(ServerCFS) Uncaught CORBA exception" << endl;
	    }
	}
 
	void run() {
	    this->orb->run();
	}

	void shutdown() {
	    cout << "Received request for shutdown. Removing the exportings..." << endl;
	    while (!exports.empty()) {
		char *path = exports.back();
		exports.pop_back();
		cout << "Removing " << path << " ... ";
		
		int error = superblock->unlink(path);
		
		if (!error) 
		    cout << "done." << endl;
		else cout << "not there!" << endl;
	    }
	}
	  
	private:
	CORBA::ORB_var orb;
	vector<char*> exports;
	Superblock_var superblock;
    };
};

CFS::ServerCFS_impl * server;

void terminate(int signal) {
    if (signal == SIGTERM || signal == SIGKILL || signal == SIGABRT) 
	server->shutdown();
    exit(0);
}

int main (int argc, char *argv[]) {
    daemon(TRUE, TRUE);
    // Create an oject
    server = new CFS::ServerCFS_impl(CORBA::ORB_init(argc, argv));
    signal(SIGTERM, terminate);
    signal(SIGKILL, terminate);
    signal(SIGABRT, terminate);
    server->run();

    return 0;
}
