/*
 * CFS (Corba File System) for Linux.
 * Livio Soares <livio@ime.usp.br>
 * http://www.lcpd.ime.usp.br/~livio/CFS/
 *    
 *  This was adapted from CODA's upcall.c to become CFS's upcall
 */

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/pagemap.h>  /* for kmap() */
/* #include <linux/dcache.h> */
/* #include <asm/statfs.h> */

#include <linux/cfs.h>

extern struct dentry_operations cfs_dentry_operations; /* for use in ll_cfs_readdir */

static int cfs_upcall(struct super_block *sb, unsigned long upcall_number)
{
	return 0;
}

int ll_cfs_read_inode(struct inode *inode, char *inodename);

/* init_structs()
 *
 * Initialize the downcall and upcall structures.
 */
#define init_calls(up, down, t)                                                        \
do {                                                                                    \
	CFS_ALLOC(up, struct cfs_upcall *, sizeof(struct cfs_upcall));                  \
	CFS_ALLOC(down, struct cfs_downcall *, sizeof(struct cfs_downcall));            \
	up->type = t;                                                                   \
	down->type = t;                                                                 \
} while (0)

/* write full pathname into buffer and return start of pathname */
static char * cfs_d_path(struct dentry *dentry, char *buf, int buflen)
{
/*	char *res;
	struct vfsmount *rootmnt;
	struct dentry *root;
	struct vfsmount *vfsmnt;

	vfsmnt = mntget(current->fs->pwdmnt);
	rootmnt = mntget(vfsmnt);
	root = dget(vfsmnt->mnt_root);
	spin_lock(&dcache_lock);
	res = __d_path(dentry, vfsmnt, root, rootmnt, buf, buflen);
	spin_unlock(&dcache_lock);
	dput(root);
	mntput(rootmnt);
	mntput(vfsmnt);
	if (!IS_ROOT(dentry) && list_empty(&dentry->d_hash))
	res[strlen(res)-10] = 0;   removes " (deleted)" from the string

	return buf; */

	char * end = buf+buflen;
	char * retval;
	int namelen;

	*--end = '\0';
	buflen--;

	/* Get '/' right */
	retval = end-1;
	*retval = '/';	

	for (;;) {
		struct dentry * parent;
		
		if (IS_ROOT(dentry))
			goto global_root;

		parent = dentry->d_parent;
		namelen = dentry->d_name.len;
		buflen -= namelen + 1;
		if (buflen < 0)
			break;
		end -= namelen;
		memcpy(end, dentry->d_name.name, namelen);
		*--end = '/';
		retval = end;
		dentry = parent;
	}
	return retval;
global_root:
	namelen = dentry->d_name.len;
	buflen -= namelen;
	if (buflen >= 0) {
		retval -= namelen-1;	/* hit the slash */
		memcpy(retval, dentry->d_name.name, namelen);
	}
	return retval;

}

/* the upcalls */
int ll_cfs_statfs(struct super_block *sb, struct statfs *buf)
{
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;

	printk("CFS: ll_cfs_statfs\n");

	init_calls(up, down, CFS_STATFS);
	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

	memcpy(buf, &(down->u.statfs), sizeof(down->u.statfs));

	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down, sizeof(struct cfs_downcall));
	return 0;
}

int ll_cfs_lookup(struct inode **inode, struct dentry *dentry)
{
	char *tmp, buf[CFS_NAME_LEN];
	int error = 0, new = 0;

	printk("CFS: ll_cfs_lookup ... ");

	if (!*inode) { /* || (!IS_ROOT(dentry) && list_empty(&dentry->d_hash))) { */
		new = 1;
		*inode = cfs_get_inode(dentry->d_sb, 0, NODEV);
	}

	if (*inode) {
		tmp = cfs_d_path(dentry, buf, CFS_NAME_LEN);
		if ((error = ll_cfs_read_inode(*inode, tmp))) {
			if (new) {
				iput(*inode);
				*inode = NULL;
			}
			return error;
		}
		cfs_get_inode_define_mode(*inode);
	}

	return error;
}

/*
 * Read a page. Again trivial. If it didn't already exist
 * in the page cache, it is zero-filled.
 */
int ll_cfs_readpage(struct dentry *dentry, loff_t offset, char * buffer)
{
	char buf[CFS_NAME_LEN], *tmp;
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;

	init_calls(up, down, CFS_READPAGE);

	tmp = cfs_d_path(dentry, buf, CFS_NAME_LEN);
	memcpy(up->u.readpage.filename, tmp, strlen(tmp)+1);

	up->u.readpage.file_offset = offset; /*((loff_t)page->index << PAGE_CACHE_SHIFT);*/

	printk("CFS: ll_cfs_readpage: file=%s, f_pos=%lld, index=%lld\n",
	       up->u.readpage.filename, offset, (loff_t)(offset/PAGE_CACHE_SIZE));

	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

	memcpy(buffer, down->u.readpage.buf, down->u.readpage.nbytes);

	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down->u.readpage.buf, down->u.readpage.nbytes);
	CFS_FREE(down, sizeof(struct cfs_downcall));
	return 0;
}

int ll_cfs_commit_write(struct file *file, struct page *page, unsigned offset, unsigned to)
{
	char buf[CFS_NAME_LEN], *tmp;
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;
	loff_t pos = ((loff_t)page->index << PAGE_CACHE_SHIFT);

	init_calls(up, down, CFS_COMMIT_WRITE);

	tmp = cfs_d_path(file->f_dentry, buf, CFS_NAME_LEN);
	memcpy(up->u.commitwrite.filename, tmp, strlen(tmp)+1);

	up->u.commitwrite.file_offset = pos + offset;
	up->u.commitwrite.nbytes = to - offset;
	up->u.commitwrite.buf = kmap(page) + /* pos + */ offset;

	printk("CFS: ll_cfs_commit_write: file_name=%s, offset=%d, to=%d, pos=%lld\n",
	       up->u.commitwrite.filename, offset, to, pos);

	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down, sizeof(struct cfs_downcall));
        return 0;
}


/*int ll_cfs_read_inode(struct inode *inode) */
int ll_cfs_read_inode(struct inode *inode, char *inodename)
{
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;

	init_calls(up, down, CFS_READ_INODE);
	memcpy(up->u.readinode.inodename, inodename, strlen(inodename)+1);
	printk("CFS: ll_cfs_read_inode: i_ino=%ld, name=%s\n", inode->i_ino, up->u.readinode.inodename);
	
	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

/*	printk("CFS: ll_cfs_read_inode: ANSWER:
dev=%d
inode=%ld
mode=%d
nlink=%d
uid=%d
gid=%d
rdev=%d
size=%ld
blksize=%ld
blocks=%ld
atime=%ld
mtime=%ld
ctime=%ld
", down->u.stat.st_dev, down->u.stat.st_ino, down->u.stat.st_mode, down->u.stat.st_nlink, 
   down->u.stat.st_uid, down->u.stat.st_gid, down->u.stat.st_rdev, down->u.stat.st_size, 
   down->u.stat.st_blksize, down->u.stat.st_blocks, down->u.stat.st_atime,
   down->u.stat.st_mtime, down->u.stat.st_ctime); */
	 
	inode->i_dev = to_kdev_t(down->u.stat.st_dev);
/*	inode->i_ino = down->u.stat.st_ino;  */
	inode->i_mode = down->u.stat.st_mode;
/*	inode->i_nlink = down->u.stat.st_nlink; */
	inode->i_uid = down->u.stat.st_uid;
	inode->i_gid = down->u.stat.st_gid;
	inode->i_rdev = to_kdev_t(down->u.stat.st_rdev);
	inode->i_size  = down->u.stat.st_size;
	inode->i_atime = down->u.stat.st_atime;
	inode->i_mtime = down->u.stat.st_mtime;
	inode->i_ctime = down->u.stat.st_ctime;

	if (down->u.stat.st_blksize) {
		inode->i_blocks = down->u.stat.st_blocks;
		inode->i_blksize = down->u.stat.st_blksize;
	}
	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down, sizeof(struct cfs_downcall));
	return 0;
}

void ll_cfs_force_delete(struct inode *inode)
{
	printk("CFS: ll_cfs_force_delete\n");
	cfs_upcall(inode->i_sb, CFS_FORCE_DELETE);
	return;
}

/*
 * File creation. Allocate an inode, and we're done..
 */
int ll_cfs_mknod(struct inode *dir, struct dentry *dentry, int mode, int dev)
{
	char buf[CFS_NAME_LEN], *tmp;
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;

	init_calls(up, down, CFS_MKNOD);
	up->u.mknod.mode = mode;
	up->u.mknod.dev = dev;
	tmp = cfs_d_path(dentry, buf, CFS_NAME_LEN);
	memcpy(up->u.mknod.inodename, tmp, strlen(tmp)+1);

	printk("CFS: ll_cfs_mknod: name=%s, mode=%o\n", up->u.mknod.inodename, mode);

	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down, sizeof(struct cfs_downcall));
	return 0;
}

int ll_cfs_mkdir(struct inode * dir, struct dentry * dentry, int mode)
{
	printk("CFS: ll_cfs_mkdir\n");
	cfs_upcall(dir->i_sb, CFS_MKDIR);
	return 0;
}

int ll_cfs_create(struct inode *dir, struct dentry *dentry, int mode)
{
	printk("CFS: ll_cfs_create\n");
	cfs_upcall(dir->i_sb, CFS_CREATE);
	return 0;
}

/*
 * Link a file..
 */
int ll_cfs_link(struct dentry *old_dentry, struct inode * dir, struct dentry * dentry)
{
	printk("CFS: ll_cfs_link\n");
	cfs_upcall(dir->i_sb, CFS_LINK);
	return 0;
}

inline int ll_cfs_positive(struct dentry *dentry)
{
	cfs_upcall(dentry->d_sb, CFS_POSITIVE);
	return 0;
}

int ll_cfs_unlink(struct inode * dir, struct dentry *dentry)
{
	char buf[CFS_NAME_LEN], *tmp;
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;

	init_calls(up, down, CFS_UNLINK);
	up->u.unlink.mode = dentry->d_inode->i_mode;
	tmp = cfs_d_path(dentry, buf, CFS_NAME_LEN);
	memcpy(up->u.unlink.inodename, tmp, strlen(tmp)+1);
	

	printk("CFS: ll_cfs_unlink: name=%s\n", up->u.unlink.inodename);

	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down, sizeof(struct cfs_downcall));
	return 0;
}

int ll_cfs_rmdir(struct inode * dir, struct dentry *dentry)
{
	printk("CFS: ll_cfs_rmdir\n");
	cfs_upcall(dir->i_sb, CFS_RMDIR);
	return 0;
}


int ll_cfs_rename(struct inode * old_dir, struct dentry *old_dentry, struct inode * new_dir,struct dentry *new_dentry)
{
	char buf[CFS_NAME_LEN], *tmp;
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;

	init_calls(up, down, CFS_RENAME);
	tmp = cfs_d_path(old_dentry, buf, CFS_NAME_LEN);
	memcpy(up->u.rename.oldpath, tmp, strlen(tmp)+1);
	tmp = cfs_d_path(new_dentry, buf, CFS_NAME_LEN);
	memcpy(up->u.rename.newpath, tmp, strlen(tmp)+1);

	printk("CFS: ll_cfs_rename: oldname=%s, newname=%s\n", up->u.rename.oldpath, up->u.rename.newpath);

	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down, sizeof(struct cfs_downcall));
	return 0;
}

int ll_cfs_symlink(struct inode * dir, struct dentry *dentry, const char * symname)
{
	char buf[CFS_NAME_LEN], *tmp;
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;

	init_calls(up, down, CFS_SYMLINK);
	tmp = cfs_d_path(dentry, buf, CFS_NAME_LEN);
	memcpy(up->u.symlink.newpath, tmp, strlen(tmp)+1);
	memcpy(up->u.symlink.oldpath, symname, strlen(symname)+1);

	printk("CFS: ll_cfs_symlink: oldname=%s, newname=%s\n", up->u.symlink.oldpath, up->u.symlink.newpath);

	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down, sizeof(struct cfs_downcall));
	return 0;
}

int ll_cfs_readlink(struct dentry *dentry, char *page, int *len)
{
	char buf[CFS_NAME_LEN], *tmp;
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;

	init_calls(up, down, CFS_READLINK);

	tmp = cfs_d_path(dentry, buf, CFS_NAME_LEN);
	memcpy(up->u.readlink.linkname, tmp, strlen(tmp)+1);

	printk("CFS: ll_cfs_readlink: link=%s\n", up->u.readlink.linkname);

	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

/*	printk("CFS: ll_cfs_readpage: nbytes=%d, file contents=%sEOF\n", 
	down->u.readpage.nbytes, down->u.readpage.buf); */

	memcpy(page, down->u.readlink.buf, down->u.readlink.nbytes);
	*len = down->u.readlink.nbytes;

	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down->u.readpage.buf, down->u.readpage.nbytes);
	CFS_FREE(down, sizeof(struct cfs_downcall));
	return 0;

}


int ll_cfs_sync_file(struct file * file, struct dentry *dentry, int datasync)
{
	printk("CFS: ll_cfs_sync_file\n");
	cfs_upcall(file->f_dentry->d_sb, CFS_SYNC_FILE);
	return 0;
}

int ll_cfs_readdir(struct file * filp, void * dirent, filldir_t filldir)
{
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error, i, j, nbytes = 0, fpos = filp->f_pos;
	char buf[CFS_NAME_LEN], *tmp;
	struct dirent* dirp;
	struct dentry *f_dentry = filp->f_dentry;
	
	tmp = cfs_d_path(f_dentry, buf, CFS_NAME_LEN);
	printk("CFS: FORA: ll_cfs_readdir: nome=%s, pos=%lld\n", tmp, filp->f_pos);

	printk("CFS: DENTRO: ll_cfs_readdir: nome=%s, pos=%lld\n", tmp, filp->f_pos);
	init_calls(up, down, CFS_READDIR);
	memcpy(up->u.readdir.dirname, tmp, strlen(tmp)+1);
	up->u.readdir.file_offset = filp->f_pos;

	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;
	
	j = 0;
	i = down->u.readdir.ndir;
	dirp = down->u.readdir.dirp;
	/* Damn! fpos came set :-/  We have to skip some entries... */
	if (filp->f_pos) 
		for (j = 0; i > 0 && j < filp->f_pos;) {
			i -= dirp->d_reclen;
			j ++;
			dirp = (struct dirent *)((char *)dirp + dirp->d_reclen);
		}

	/* Register the inodes here? First, get them from the server.
	   Second, register in the dcache. */
	for (; i > 0; j++) {
		if (j == filp->f_pos) {
			if (filldir(dirent, dirp->d_name, strlen(dirp->d_name), 
				    filp->f_pos, dirp->d_ino, DT_UNKNOWN) < 0)
				break;
			
			printk("CFS: ll_cfs_readdir: [%d].[%d]=%s father=%s next=%d\n", j, i,
			       dirp->d_name, f_dentry->d_name.name, dirp->d_reclen);
			
			filp->f_pos++;
			nbytes += dirp->d_reclen;
		}
		i -= dirp->d_reclen;
		dirp = (struct dirent *)((char *)dirp + dirp->d_reclen);
	}
	
	CFS_FREE(down->u.readdir.dirp, down->u.readdir.ndir);
	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down, sizeof(struct cfs_downcall)); 

	return nbytes;
/*	return 0; */
}

int ll_cfs_setattr(struct dentry *dentry, struct iattr * attr)
{
	char buf[CFS_NAME_LEN], *tmp;
	struct cfs_upcall *up;
	struct cfs_downcall *down;
	int error;

	printk("CFS: ll_cfs_setattr\n");

	init_calls(up, down, CFS_SETATTR);
	tmp = cfs_d_path(dentry, buf, CFS_NAME_LEN);
	memcpy(up->u.setattr.inodename, tmp, strlen(tmp)+1);
	memcpy(&(up->u.setattr.attr), attr, sizeof(struct iattr));

	if ((error = psdev_enqueue(up, down)) < 0)
		return error;
	if (down->error < 0)
		return down->error;

	CFS_FREE(up, sizeof(struct cfs_upcall));
	CFS_FREE(down, sizeof(struct cfs_downcall));
	return 0;
}

struct super_block *ll_cfs_read_super(struct super_block * sb, void * data, int silent)
{
	printk("CFS: ll_cfs_read_super\n");
	cfs_upcall(sb, CFS_READ_SUPER);
	return NULL;
}

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
